#include <stdio.h>
#include <stdint.h>
#include <assert.h>
#include "det.h"

// input: array s[0],s[1],...,s[n-1]
// defining S = sum_i s[i]*65536^i
// output: overwriting s
// defining -S = sum_i news[i]*65536^i
// aborts on overflow
static void int16array_negate(int16_t *s,long long n)
{
  long long carry = 0;
  for (long long i = 0;i < n;++i) {
    assert(carry >= 0);
    assert(carry <= 1);
    carry -= s[i];
    assert(-32768 <= carry);
    assert(carry <= 32769);
    s[i] = carry;
    carry = carry>=32768;
  }
  assert(carry == 0);
}

// input: negate is 0 or 1
// output: array s[0],s[1],...,s[n-1]
// defining z = sum_i s[i]*65536^i
// aborts on overflow
static void int16array_from_mpz(int16_t *s,long long n,const mpz_t z)
{
  long long i;

  assert((long long) mpz_sizeinbase(z,2) <= 16*n-2);
  // i.e. |z| < 2^(16n-2)

  for (i = 0;i < n;++i) s[i] = 0;
  mpz_export(s,0,-1,1,0,0,z);
  // now |z| is sum_i ((uint16_t *)s)[i]*65536^i

  long long carry = 0;
  for (long long i = 0;i < n;++i) {
    assert(0 <= carry);
    assert(carry <= 1);
    carry += ((uint16_t *) s)[i];
    assert(0 <= carry);
    assert(carry <= 65536);
    s[i] = carry;
    carry = carry>=32768;
  }
  assert(carry == 0);

  // now |z| is sum_i s[i]*65536^i
  if (mpz_sgn(z) < 0) int16array_negate(s,n);
}

// input: array s[0],s[1],...,s[n-1]
// output: z = sum_i s[i]*65536^i mod 65536^n+1
// overwriting t[0],t[1],...,t[n-1]
static void mpz_from_int16array(mpz_t z,const int16_t *s,long long n,uint16_t *t)
{
  // define S = sum_i s[i]*65536^i

  long long carry = 0;
  for (long long i = 0;i < n;++i) {
    assert(-1 <= carry);
    assert(carry <= 0);
    carry += s[i];
    assert(-32769 <= carry);
    assert(carry <= 32767);
    t[i] = carry;
    carry >>= 16;
  }
  assert(-1 <= carry);
  assert(carry <= 0);

  // S is sum_i t[i]*65536^i + carry*65536^n

  mpz_import(z,2*n,-1,1,0,0,t);
  // S is z + carry*65536^n
  // so S is z-carry mod 65536^n+1

  if (carry) mpz_add_ui(z,z,1);
}

static void cyclo2power_det_signed_optreal(mpz_t det,const mpz_t *f,long long n,long long real)
{
  assert(n >= 1);
  assert(!(n&(n-1)));

  if (n == 1 || (n == 2 && real)) {
    mpz_set(det,f[0]);
    return;
  }

  long long nbits = 0;
  while ((n >> nbits) > 1) ++nbits;
  assert(n == (1LL<<nbits));
  assert(nbits >= 1); // since n >= 2

  // ----- allocations begin here

  mpz_t maxabsfi;
  mpz_init(maxabsfi);
  mpz_set_ui(maxabsfi,0);

  mpz_t absfi;
  mpz_init(absfi);

  if (real)
    for (long long i = 1;i < n;++i) {
      mpz_neg(absfi,f[i]);
      assert(mpz_cmp(f[n-i],absfi) == 0);
    }
  for (long long i = 0;i < n;++i) {
    mpz_abs(absfi,f[i]);
    if (mpz_cmp(absfi,maxabsfi) > 0)
      mpz_set(maxabsfi,absfi);
  }
  mpz_clear(absfi);

  long long maxbits = mpz_sizeinbase(maxabsfi,2)+1;
  // now 2|f_i| < 2^maxbits
  // XXX: can cut this closer
  // but beware that |f_i| < 2^maxbits is not good enough for signed representation

  mpz_clear(maxabsfi);

  long long r = (nbits+maxbits+15)/16;
  long long W = r*n;

  // nbits+maxbits <= 16r
  // in particular maxbits <= 16r-1
  // so 2|f_i| < 2^maxbits <= 2^(16r-1)

  mpz_t modulus;
  mpz_init(modulus);
  mpz_set_ui(modulus,1);
  mpz_mul_2exp(modulus,modulus,16*W);
  mpz_add_ui(modulus,modulus,1);
  // modulus is 2^(16W)+1

  mpz_t halfmodulus;
  mpz_init(halfmodulus);
  mpz_set_ui(halfmodulus,1);
  mpz_mul_2exp(halfmodulus,halfmodulus,16*W-1);
  // halfmodulus is 2^(16W-1)

  mpz_t minushalfmodulus;
  mpz_init(minushalfmodulus);
  mpz_neg(minushalfmodulus,halfmodulus);
  // minushalfmodulus is -2^(16W-1); not used except for assert

  int16_t uwords[W];
  uint16_t tmp[W];
  mpz_t u[2];
  mpz_init(u[0]);
  mpz_init(u[1]);

  for (long long i = 0;i < n;++i)
    int16array_from_mpz(uwords+r*i,r,f[i]);
  mpz_from_int16array(u[0],uwords,W,tmp);
  if (mpz_cmp(u[0],halfmodulus) > 0)
    mpz_sub(u[0],u[0],modulus);

  for (;;) {
    // loop invariants:
    // define f_j = sum_{0<=i<r} uwords[rj+i]2^16i
    // then u[0] = sum_i uwords[i] 2^16i = sum_j f_j 2^16rj
    // also 2|f_i| < 2^maxbits <= 2^(16r-1)
    assert(nbits+maxbits <= 16*r);
    assert(maxbits <= 16*r-1);
    assert(n == (1LL<<nbits));
    assert(W == r*n);

    if (r == W) break;
    if (real) if (r == W/2) break;

    for (long long j = r;j < W;j += 2*r)
      int16array_negate(uwords+j,r);
    mpz_from_int16array(u[1],uwords,W,tmp);
    if (mpz_cmp(u[1],halfmodulus) > 0)
      mpz_sub(u[1],u[1],modulus);
    assert(mpz_cmp(u[1],halfmodulus) <= 0);
    assert(mpz_cmp(u[1],minushalfmodulus) >= 0);
    
    // now u[1] = sum_j (-1)^j f_j 2^16rj

    mpz_mul(u[0],u[0],u[1]);
    // XXX: take advantage of the 2^16W+1 modulus for faster multiplications

    // now u[0] is sum_j g_j 2^32rj
    // where g(x^2) = f(x)f(-x)

    // reduce mod 2^16W+1:
    mpz_fdiv_q_2exp(u[1],u[0],16*W); // at most 2^(16W-2) in absolute value
    mpz_fdiv_r_2exp(u[0],u[0],16*W); // between 0 and 2^16W
    mpz_sub(u[0],u[0],u[1]); // slightly wider range than 0...2^16W
    if (mpz_cmp(u[0],halfmodulus) > 0)
      mpz_sub(u[0],u[0],modulus);
    assert(mpz_cmp(u[0],halfmodulus) <= 0);
    assert(mpz_cmp(u[0],minushalfmodulus) >= 0);

    // each g coeff is a sum/diff of n products of f coeffs
    // each product of f coeffs has abs value < 2^(2*maxbits-2)
    // so each g coeff has abs value < 2^(nbits+2*maxbits-2)
    // and nbits+maxbits <= 16r and maxbits <= 16r-1
    // implying 2*nbits+2*maxbits <= 32r and nbits+2*maxbits <= 32r-1

    r *= 2;
    maxbits = maxbits*2+nbits; // not used except for assert
    nbits -= 1; // not used except for assert
    n /= 2; // not used except for assert

    // now u[0] is sum_i g_i 2^16ri
    // and each g coeff has abs value < 2^(maxbits-1)
    // and nbits+maxbits <= 16r and maxbits <= 16r-1

    // XXX: usually possible to reduce r somewhat
    // but this requires taking time to check and to reorganize data

    // XXX: once r is big, can align to 32, 64, etc. for faster loops

    int16array_from_mpz(uwords,W,u[0]);
  }

  mpz_set(det,u[0]);

  mpz_clear(u[0]);
  mpz_clear(u[1]);
  mpz_clear(modulus);
  mpz_clear(halfmodulus);
  mpz_clear(minushalfmodulus);
}

// input n: power of 2
// input f: array of n integers
// output det: det of F from \Z[x]/(x^n+1) down to \Z 
// i.e. resultant(x^n+1,F)
// where F = f[0]+f[1]x+...+f[n-1]x^(n-1)
void cyclo2power_det_signed(mpz_t det,const mpz_t *f,long long n)
{
  cyclo2power_det_signed_optreal(det,f,n,0);
}

// input n: power of 2
// input f: array of n integers
// satisfying f[i] = -f[n-i] for 1<=i<=n-1
// output det: det of F from real subring of \Z[x/(x^n+1) down to \Z
void cyclo2power_det_signed_real(mpz_t det,const mpz_t *f,long long n)
{
  cyclo2power_det_signed_optreal(det,f,n,1);
}

void cyclo2power_det_abs(mpz_t det,const mpz_t *f,long long n)
{
  cyclo2power_det_signed(det,f,n);
  mpz_abs(det,det);
}

void cyclo2power_det_abs_real(mpz_t det,const mpz_t *f,long long n)
{
  cyclo2power_det_signed_real(det,f,n);
  mpz_abs(det,det);
}
